!###############################################################################
!
#define IF_NOTOK_RETURN(action) if (status/=0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
!
!###############################################################################

module tmm_mf

  use GO  , only : gol, goErr, goPr
  use GO  , only : TDate
  use Grid, only : TllGridInfo, TggGridInfo, TshGridInfo, TshGrid, TLevelInfo

#include "tmm.h"

#ifdef with_hdf  
  use tmm_mf_hdf  , only : TMeteoFile_hdf
#endif
#ifdef with_grib
  use tmm_mf_grib , only : TMeteoFile_grib
#endif
#ifdef with_ncep
  use tmm_mf_ncep_cdc, only : TMeteoFile_ncep_cdc
  use tmm_mf_ncep_gfs, only : TMeteoFile_ncep_gfs
#endif
#ifdef with_prism  
  use tmm_mf_prism, only : TMeteoFile_prism
#endif
#ifdef with_msc
  use tmm_mf_msc  , only : TMeteoFile_msc
#endif
  
  implicit none
  
  ! --- in/out -------------------------------------
   
  private
 
  public   ::  TMeteoFile
  public   ::  Init, Done
  public   ::  Opened, CheckTime, CheckParam

  public   ::  SetupInput
  public   ::  ReadRecord
!  public   ::  ReadEqvLatStuff

  public   ::  SetupOutput
  public   ::  WriteRecord
  
  
  ! --- const ---------------------------------------
  
  character(len=*), parameter  ::  mname = 'tmm_mf'

  ! --- types --------------------------------------
  
  type TMeteoFile
    ! opened yet ?
    logical                    ::  opened
    character(len=1)           ::  io=''
    ! meteo archive keys:
    character(len=256)         ::  dir=''
    character(len=256)         ::  archivekey=''
    ! parameter keys for fields in this file
    character(len=256)         ::  paramkeys=''
    ! time range for which file is valid
    type(TDate)                ::  t1, t2
    !
    ! access to current meteo file
    !
    character(len=10)          ::  filetype=''
    character(len=256)         ::  filename=''
    character(len=256)         ::  spm_filename=''
#ifdef with_hdf
    type(TMeteoFile_hdf)       ::  mf_hdf         ! hdf file
#endif
#ifdef with_grib
    type(TMeteoFile_grib)      ::  mf_grib        ! grib file
#endif
#ifdef with_ncep  
    type(TMeteoFile_ncep_cdc)  ::  mf_ncep_cdc    ! ncep file from cdc archive
    type(TMeteoFile_ncep_gfs)  ::  mf_ncep_gfs    ! ncep gfs file
#endif
#ifdef with_prism  
    type(TMeteoFile_prism)     ::  mf_prism       ! prism file
#endif
#ifdef with_msc
    type(TMeteoFile_msc)       ::  mf_msc
#endif
  end type TMeteoFile


  ! --- interfaces -------------------------------
  
  interface Init
    module procedure mf_Init
  end interface

  interface Done
    module procedure mf_Done
  end interface

  interface Opened
    module procedure mf_Opened
  end interface

  interface CheckTime
    module procedure mf_CheckTime
  end interface

  interface CheckParam
    module procedure mf_CheckParam
  end interface

  interface SetupInput
    module procedure mf_SetupInput
  end interface

  interface ReadRecord
    module procedure mf_ReadRecord
  end interface

!  interface ReadEqvLatStuff
!    module procedure mf_ReadEqvLatStuff
!  end interface

  interface SetupOutput
    module procedure mf_SetupOutput
  end interface

  interface WriteRecord
    module procedure mf_WriteRecord_2d
    module procedure mf_WriteRecord_3d
  end interface


contains


  ! ===========================================================
  !
  !   init/done
  !
  ! ===========================================================
  

  subroutine mf_Init( mf, io, status )
  
    ! --- begin -------------------------------------------
    
    type(TMeteoFile), intent(out)           ::  mf
    character(len=1), intent(in)            ::  io
    integer, intent(out)                    ::  status
    
    ! --- const -------------------------------------------
    
    character(len=*), parameter ::  rname = mname//'/mf_Init'
    
    ! --- begin -------------------------------------------
    
    ! input or output ?
    mf%io = io

    ! file not opened yet
    mf%opened = .false.
    
    ! ok
    status = 0
    
  end subroutine mf_Init
  
  
  ! ***
  
  
  subroutine mf_Done( mf, status )
  
    use GO, only : goSystem
    use Grid, only : Done

#ifdef with_hdf
    use tmm_mf_hdf      , only : Done
#endif
#ifdef with_grib
    use tmm_mf_grib     , only : Done
#endif
#ifdef with_ncep
    use tmm_mf_ncep_cdc , only : Done
    use tmm_mf_ncep_gfs , only : Done
#endif
#ifdef with_prism  
    use tmm_mf_prism    , only : Done
#endif
#ifdef with_msc
    use tmm_mf_msc  , only : Done
#endif
  
    ! --- begin -------------------------------------------
    
    type(TMeteoFile), intent(inout)         ::  mf
    integer, intent(out)                    ::  status
    
    ! --- const -------------------------------------------
    
    character(len=*), parameter ::  rname = mname//'/mf_Done'
    
    ! --- begin -------------------------------------------

    ! close file if necessary
    if ( mf%opened ) then
      select case ( mf%filetype )
#ifdef with_hdf  
        case ( 'hdf' )
          call Done( mf%mf_hdf, status )
          IF_NOTOK_RETURN(status=1)
#endif
#ifdef with_grib  
        case ( 'grib' )
          call Done( mf%mf_grib, status )
          IF_NOTOK_RETURN(status=1)
#endif
#ifdef with_ncep  
        case ( 'ncep-cdc' )
          call Done( mf%mf_ncep_cdc, status )
          IF_NOTOK_RETURN(status=1)
        case ( 'ncep-gfs' )
          call Done( mf%mf_ncep_gfs, status )
          IF_NOTOK_RETURN(status=1)
#endif
#ifdef with_prism  
        case ( 'prism' )
          call Done( mf%mf_prism, status )
          IF_NOTOK_RETURN(status=1)
#endif
#ifdef with_msc
        case ( 'msc-data' )
          call Done( mf%mf_msc, status )
          IF_NOTOK_RETURN(status=1)
#endif
        case default
          write (gol,'("unsupported filetype `",a,"`")') trim(mf%filetype); call goErr
          write (gol,'("in ",a)') rname; call goErr; status=1; return
      end select
      mf%opened = .false.
    end if
    
    ! ok
    status = 0
    
  end subroutine mf_Done
  
  
  ! ***
  
  
  logical function mf_Opened( mf )

    ! --- begin -------------------------------------------
    
    type(TMeteoFile), intent(in)         ::  mf
    
    ! --- begin -------------------------------------------
    
    mf_Opened = mf%opened
    
  end function mf_Opened
    

  ! ===========================================================
  !
  !   check contents of open meteo file
  !
  ! ===========================================================
  

  ! Check time in meteo file;
  ! status: 
  !   <0    : mf does not include [t1,t2]
  !    0    : mf includes [t,t2]
  !   >0    : error; mf not open ?
  !       
  
  subroutine mf_CheckTime( mf, t1, t2, status )

    use GO, only : TDate, IncrDate, wrtgol, IsAnyDate
    use GO, only : operator(+), operator(-), operator(==), operator(<), operator(<=)
   
    ! --- begin -------------------------------------------
    
    type(TMeteoFile), intent(in)           ::  mf
    type(TDate), intent(in)                ::  t1, t2
    integer, intent(out)                   ::  status
    
    ! --- const -------------------------------------------
    
    character(len=*), parameter ::  rname = mname//'/mf_CheckTime'
    
    ! --- begin -------------------------------------------

    ! not open ?
    if ( .not. Opened(mf) ) then
      write (gol,'("meteo file not opened")'); call goErr
      write (gol,'("in ",a)') rname; call goErr; status = 1; return
    end if
    
    ! trap any date:
    if ( IsAnyDate(t1) .and. IsAnyDate(t2) ) then
      status =  0; return
    end if
    
    ! [t1,t2] is either:
    !   covered by mf   ->  status =  0     
    !   older than mf   ->  status = -2
    !   newer than mf   ->  status = -1
    !   error ... (half in, half outside mf)
    ! seperate tests for intervals and instant time:
    if ( t1 == t2 ) then

      ! instant time
      if ( ( (mf%t1 <= t1) .and. (t1 <= mf%t2) ) ) then
        status =  0; return
      else if ( t1 < mf%t1 ) then
        status = -2; return
      else if ( mf%t2 <= t1 ) then
        status = -1; return
      else
        write (gol,'("requested instant time t1 (=t2) overlaps part of mf time:")'); call goErr
        call wrtgol( 'ERROR -   t1     : ', t1 ); call goErr
        call wrtgol( 'ERROR -   t2     : ', t2 ); call goErr
        call wrtgol( 'ERROR -   mf%t1  : ', mf%t1 ); call goErr
        call wrtgol( 'ERROR -   mf%t2  : ', mf%t2 ); call goErr
        write (gol,'("  params : ",a)') trim(mf%paramkeys); call goErr
        write (gol,'("in ",a)') rname; call goErr; status = 1; return
      end if

    else if ( t1 < t2 ) then

      ! interval
      ! extra: [t1,t2] is covered by mf%(t1,t2) ...
      if ( ( (mf%t1                  <= t1) .and. (t2 <= mf%t2                 ) ) .or. &
           ( (mf%t1-IncrDate(mili=1) <= t1) .and. (t2 <= mf%t2+IncrDate(mili=1)) )      ) then
        status =  0; return
      else if ( t2 <= mf%t1 ) then
        ! request for field older than those in file
        status = -2; return
      else if ( mf%t2 <= t1 ) then
        ! request for field newer than those in file
        status = -1; return
      else
        write (gol,'("requested interval [t1,t2] overlaps part of mf time:")'); call goErr
        call wrtgol( 'ERROR -   t1     : ', t1 ); call goErr
        call wrtgol( 'ERROR -   t2     : ', t2 ); call goErr
        call wrtgol( 'ERROR -   mf%t1  : ', mf%t1 ); call goErr
        call wrtgol( 'ERROR -   mf%t2  : ', mf%t2 ); call goErr
        write (gol,'("  params : ",a)') trim(mf%paramkeys); call goErr
        write (gol,'("in ",a)') rname; call goErr; status = 1; return
      end if

    else 
    
      write (gol,'("arguments should specify an instant time or valid interval :")'); call goErr
      call wrtgol( 'ERROR -   t1 : ', t1 ); call goErr
      call wrtgol( 'ERROR -   t2 : ', t2 ); call goErr
      write (gol,'("in ",a)') rname; call goErr; status = 1; return
    
    end if
      
    ! something wrong if this point is reached ...
    status = 1
    
  end subroutine mf_CheckTime
    
  
  ! ***
  
  
  ! Check if param is included in meteo file;
  ! status: 
  !   <0    : mf does not include param
  !    0    : mf includes param
  !   >0    : error; mf not open ?
  !       
  
  subroutine mf_CheckParam( mf, io, archivekey, paramkey, status )
  
    use GO, only : goLoCase

    ! --- begin -------------------------------------------
    
    type(TMeteoFile), intent(in)           ::  mf
    character(len=*), intent(in)           ::  io
    character(len=*), intent(in)           ::  archivekey
    character(len=*), intent(in)           ::  paramkey
    integer, intent(out)                   ::  status
    
    ! --- const -------------------------------------------
    
    character(len=*), parameter ::  rname = mname//'/mf_CheckParam'
    
    ! --- local --------------------------------------------
    
    integer             ::  pos
    
    ! --- begin -------------------------------------------
    
    ! not open ?
    if ( .not. Opened(mf) ) then
      write (gol,'("meteo file not opened")'); call goErr
      write (gol,'("in ",a)') rname; call goErr; status = 1; return
    end if
    
    ! by default not found ..
    status = -1
    
    ! wrong input/output ? then leave:
    if ( io /= mf%io ) return
    
    ! wrong grid ? then leave
    if ( archivekey /= mf%archivekey ) return
    
    ! param list is for example: '-ps-pu-pv-',
    ! thus search for example for '-pu-' ...
    ! convert all to lowercase
    pos = index( goLoCase(mf%paramkeys), '-'//trim(goLoCase(paramkey))//'-' )
    if ( pos < 1 ) return
    
    ! ok
    status = 0

  end subroutine mf_CheckParam
    
    
    
  ! ===========================================================
  !
  !   open meteo file for input
  !
  ! ===========================================================
  

  !
  ! Open the meteo file that contains the field specified by
  ! archivekey, parameter key, time,
  ! or do nothing if the requested file has been opened already.
  !
  ! <archivekey> = <archivetype>:<archivename>
  !
  !    tmpp:od-fc-ml60-glb3x2
  !   tmppS:od-fc-ml60-glb3x2
  !    grib:od-fc-ml60-glb3x2
  !   prism:
  !
  
  subroutine mf_SetupInput( mf, archivekey, paramkey, tday, t1, t2, &
                                rcfilename, dir, status )

    use GO, only : goSplitLine, goReadFromLine
    use GO, only : goSystem
    use GO, only : TrcFile, Init, Done, ReadRc
    use GO, only : TDate, IncrDate, Get, NewDate, wrtgol, &
                   Operator(+), Operator(-), Operator(/)
#ifdef with_hdf  
    use tmm_mf_hdf     , only : Init, Get
#endif
#ifdef with_grib  
    use tmm_mf_grib    , only : Init, Get
#endif
#ifdef with_ncep  
    use tmm_mf_ncep_cdc, only : Init, Get
    use tmm_mf_ncep_gfs, only : Init, Get
#endif
#ifdef with_prism  
    use tmm_mf_prism   , only : Init
#endif
#ifdef with_msc
    use tmm_mf_msc     , only : Init, Get
#endif
    
    ! --- in/out -------------------------------------

    type(TMeteoFile), intent(inout)        ::  mf
    character(len=*), intent(in)           ::  archivekey
    character(len=*), intent(in)           ::  paramkey
    type(TDate), intent(in)                ::  tday, t1, t2
    character(len=*), intent(in)           ::  rcfilename
    character(len=*), intent(in)           ::  dir
    integer, intent(inout)                 ::  status
    
    ! --- const -------------------------------------
    
    character(len=*), parameter ::  rname = mname//'/mf_SetupInput'
    
    ! name of info file:
    character(len=*), parameter  ::  infofilename = 'tmm_info.rc'
  
    ! --- local -------------------------------------
    
    character(len=10)        ::  archivetype
    character(len=256)       ::  archivename
    
    character(len=256)       ::  command
    integer                  ::  year1, month1, day1, hour1
    integer                  ::  year2, month2, day2, hour2
    integer                  ::  dth
    type(TrcFile)            ::  infofile

    character(len=256)       ::  archivename2
    character(len=10)        ::  mclass
    character(len=10)        ::  mtype
    character(len=10)        ::  mlevs
    character(len=10)        ::  mgrid
    character(len=10)        ::  filekey
    character(len=16)        ::  treskey
    logical                  ::  with_spm
    logical                  ::  constant

    ! --- begin -------------------------------------
    
    ! store archive key:
    mf%archivekey = trim(archivekey)
    mf%dir        = trim(dir)
    
    ! split archive key in type and name:
    call goSplitLine( archivekey, archivetype, ':', archivename, status )
    IF_NOTOK_RETURN(status=1)
    
    ! usually, meteo is storred in file;
    ! for PRISM project, meteo is in the memory ...
    select case ( archivetype )
    
      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      ! hdf files written by tmpp or tm5
      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifdef with_hdf

      case ( 'tmpp', 'tm5' )

        ! wich of the 'mf%mf_???' is used ?
        mf%filetype = 'hdf'

        ! setup file:
        call Init( mf%mf_hdf, 'i', dir, archivename//';form='//trim(archivetype), paramkey, &
                               tday, t1, t2, status )
        IF_NOTOK_RETURN(status=1)
        
        ! store filename:
        mf%filename = mf%mf_hdf%fname
        
        ! extract time range:
        call Get( mf%mf_hdf, status, trange1=mf%t1, trange2=mf%t2 )
        IF_NOTOK_RETURN(status=1)
        
        ! extract paramkeys for fields in file:
        call Get( mf%mf_hdf, status, paramkeys=mf%paramkeys )
        IF_NOTOK_RETURN(status=1)
          
#endif


!      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
!      ! grib files (tostie)
!      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
!
!#ifdef with_grib
!      case ( 'grib' )
!
!        ! read some settings
!        call Init( infofile, rcfilename )
!        call ReadRc( infofile, 'tmm.dir', mf%dir )
!        call Done( infofile )
!        
!        ! time resolution
!        dth = 3   ! hour
!
!        ! time range [00,24] + [-1:30,1:30] = [22:30,01:30]
!        call Get( t1, year=year1, month=month1, day=day1, hour=hour1 )
!        mf%t1 = NewDate( year=year1, month=month1, day=day1, hour=00 ) - IncrDate(hour=dth)/2
!        mf%t2 = mf%t1 + IncrDate(hour=24) + IncrDate(hour=dth)/2
!
!        !call wrtgol( '     fields valid from : ', mf%t1 )
!        !call wrtgol( '                  to   : ', mf%t2 )
!
!        ! file type and name:
!        !   od-fc-2000-01-ml60-glb1x1-T_20000111_00p03.hdf
!        ! * set type:
!        mf%filetype = 'grib'
!        ! * split archivename
!        archivename2 = archivename
!        call goReadFromLine( archivename2, mclass, sep='-' )
!        call goReadFromLine( archivename2, mtype , sep='-' )
!        call goReadFromLine( archivename2, mlevs , sep='-' )
!        call goReadFromLine( archivename2, mgrid , sep='-' )
!        ! * other
!        treskey = '00p03'
!        with_spm = .true.
!        constant = .false.
!        ! * overwrite some fields:
!        select case ( paramkey )
!          case ( 'T' )
!            filekey = 'T'
!            mf%paramkeys = '-T-'
!          case ( 'Q' )
!            filekey = 'Q'
!            mf%paramkeys= '-Q-'
!          case ( 'oro' )
!            mf%paramkeys= '-oro-'
!            mlevs = 'sfc'
!            filekey = 'oro'
!            year1  = 0
!            month1 = 0
!            day1   = 0
!            treskey = '00p00'
!            with_spm = .false.
!            constant = .true.
!          case ( 'spm' )
!            mlevs = 'sfc'
!            filekey = 'SP'
!            mf%paramkeys= '-spm-'
!            with_spm = .false.
!          case default
!            write (gol,'("unsupported grib paramkey `",a,"`")') paramkey; call goErr
!            write (gol,'("in ",a)') rname; call goErr; status=1; return
!        end select
!        ! * create filenames
!        write (mf%filename,'(a,"/",a,"-",a,"-",i4.4,"-",i2.2,"-",a,"-",a,"-",a,"_",i4.4,i2.2,i2.2,"_",a,".gb")') &
!            trim(mf%dir), &
!            trim(mclass), trim(mtype), year1, month1, trim(mlevs), trim(mgrid), &
!            trim(filekey), year1, month1, day1, trim(treskey)
!        if ( with_spm ) then
!          write (mf%spm_filename,'(a,"/",a,"-",a,"-",i4.4,"-",i2.2,"-",a,"-",a,"-",a,"_",i4.4,i2.2,i2.2,"_",a,".gb")') &
!              trim(mf%dir), &
!              trim(mclass), trim(mtype), year1, month1, 'sfc', trim(mgrid), &
!              'SP', year1, month1, day1, trim(treskey)
!        else
!          mf%spm_filename = 'none'
!        endif
!
!        ! open file(s) :
!        select case ( mf%filetype )
!          case ( 'grib' )
!            !write (*,'("    open grib file ...")')
!            call Init( mf%mf_grib, mf%filename, mf%spm_filename, constant, status )
!            IF_NOTOK_RETURN(status=1)
!          case default
!            write (gol,'("unsupported filetype `",a,"`")') trim(mf%filetype); call goErr
!            write (gol,'("  archivetype : ",a)') trim(archivetype); call goErr
!            write (gol,'("in ",a)') rname; call goErr; status=1; return
!        end select
!#endif
!
!
      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      ! grib files
      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifdef with_grib

      case ( 'ecmwf' )

        ! wich of the 'mf%mf_???' is used ?
        mf%filetype = 'grib'

        ! setup file:
        call Init( mf%mf_grib, dir, archivename, paramkey, &
                               tday, t1, t2, status )
        IF_NOTOK_RETURN(status=1)
        
        ! store filename:
        mf%filename = mf%mf_grib%fname
        
        ! extract time range:
        call Get( mf%mf_grib, status, trange1=mf%t1, trange2=mf%t2 )
        IF_NOTOK_RETURN(status=1)
        
        ! extract list of parameters in files:
        call Get( mf%mf_grib, status, paramkeys=mf%paramkeys )
        IF_NOTOK_RETURN(status=1)
          
#endif


      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      ! ncep files
      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifdef with_ncep

      case ( 'ncep-cdc' )

        ! wich of the 'mf%mf_???' is used ?
        mf%filetype = 'ncep-cdc'

        ! setup file:
        call Init( mf%mf_ncep_cdc, dir, archivename, paramkey, &
                               tday, t1, t2, status )
        IF_NOTOK_RETURN(status=1)
        
        ! store dummy filename (might be used in error messages)
        mf%filename = 'ncep cdc file'
        
        ! extract time range:
        call Get( mf%mf_ncep_cdc, status, trange1=mf%t1, trange2=mf%t2 )
        IF_NOTOK_RETURN(status=1)
        
        ! extract list of parameters in files:
        call Get( mf%mf_ncep_cdc, status, paramkeys=mf%paramkeys )
        IF_NOTOK_RETURN(status=1)
          
      case ( 'ncep-gfs' )

        ! wich of the 'mf%mf_???' is used ?
        mf%filetype = 'ncep-gfs'

        ! setup file:
        call Init( mf%mf_ncep_gfs, dir, archivename, paramkey, &
                               tday, t1, t2, status )
        IF_NOTOK_RETURN(status=1)
        
        ! store filename:
        mf%filename = mf%mf_ncep_gfs%fname
        
        ! extract time range:
        call Get( mf%mf_ncep_gfs, status, trange1=mf%t1, trange2=mf%t2 )
        IF_NOTOK_RETURN(status=1)
        
        ! extract list of parameters in files:
        call Get( mf%mf_ncep_gfs, status, paramkeys=mf%paramkeys )
        IF_NOTOK_RETURN(status=1)
          
#endif


      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      ! msc-data text files
      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifdef with_msc

      case ( 'msc-data' )

        ! wich of the 'mf%mf_???' is used ?
        mf%filetype = 'msc-data'

        ! setup file:
        call Init( mf%mf_msc, dir, archivename, paramkey, &
                               tday, t1, t2, status )
        IF_NOTOK_RETURN(status=1)
        
        ! store filename:
        mf%filename = mf%mf_msc%fname
        
        ! extract time range:
        call Get( mf%mf_msc, status, trange1=mf%t1, trange2=mf%t2 )
        IF_NOTOK_RETURN(status=1)
        
        ! extract which fields are stored in the file:
        call Get( mf%mf_msc, status, paramkeys=mf%paramkeys )
        IF_NOTOK_RETURN(status=1)
          
#endif


      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      ! prism meteo in memory
      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifdef with_prism  
      case ( 'prism' )
    
        ! wich of the 'mf%mf_???' is used ?
        mf%filetype = 'prism'

        ! setup file:
        call Init( mf%mf_prism, paramkey, status )
        IF_NOTOK_RETURN(status=1)

        ! dummy file name:
        mf%filename = 'oasis'

        ! infinite time range ...
        mf%t1 = NewDate( year=1900, month=1, day=1, hour=1 )
        mf%t2 = NewDate( year=9999, month=9, day=9, hour=9 )

        ! only the requested parameter is provided by this prism 'file' ...
        mf%paramkeys = '-'//paramkey//'-'

#endif

      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      ! error ...
      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      case default

        write (gol,'("unsupported archivetype `",a,"`")') trim(archivetype); call goErr
        write (gol,'("in ",a)') rname; call goErr; status=1; return

    end select

    ! file is opened (or, at least file name is known)
    mf%opened = .true.
    
    ! ok
    status = 0

  end subroutine mf_SetupInput
  
  

  ! ===========================================================
  !
  !   read fields, grid definition, etc
  !
  ! ===========================================================
  

  subroutine mf_ReadRecord( mf, paramkey, t1, t2, nuv, nw, &
                                gridtype, levi, &
                                lli, ll, sp_ll, &
                                ggi, gg, sp_gg, &
                                shi, sh, lnsp_sh, &
                                tmi, status )

    use GO             , only : TDate, operator(+), operator(-), operator(/)
    use Grid           , only : TllGridInfo, TggGridInfo, TshGridInfo, TLevelInfo
    use tmm_info       , only : TMeteoInfo, Init, AddHistory
#ifdef with_hdf  
    use tmm_mf_hdf     , only : ReadRecord
#endif
#ifdef with_grib  
    use tmm_mf_grib    , only : ReadRecord
#endif
#ifdef with_ncep  
    use tmm_mf_ncep_cdc, only : ReadRecord
    use tmm_mf_ncep_gfs, only : ReadRecord
#endif
#ifdef with_prism  
    use tmm_mf_prism   , only : ReadRecord
#endif
#ifdef with_msc
    use tmm_mf_msc  , only : ReadRecord
#endif

    ! --- in/out -------------------------------
    
    type(TMeteoFile), intent(inout)      ::  mf
    character(len=*), intent(in)         ::  paramkey
    type(TDate), intent(in)              ::  t1, t2
    character(len=1), intent(in)         ::  nuv
    character(len=1), intent(in)         ::  nw

    character(len=2), intent(out)        ::  gridtype
    type(TLevelInfo), intent(out)        ::  levi
    type(TllGridInfo), intent(inout)     ::  lli
    real, pointer                        ::  ll(:,:,:)
    real, pointer                        ::  sp_ll(:,:)
    type(TggGridInfo), intent(inout)     ::  ggi
    real, pointer                        ::  gg(:,:)
    real, pointer                        ::  sp_gg(:)
    type(TshGridInfo), intent(inout)     ::  shi
    complex, pointer                     ::  sh(:,:)
    complex, pointer                     ::  lnsp_sh(:)
    type(TMeteoInfo), intent(out)        ::  tmi
    integer, intent(out)                 ::  status
    
    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/mf_ReadRecord'
    
    ! --- local --------------------------------------
    
    type(TDate)         ::  tmid
    
    ! --- begin ---------------------------------
    
    !write (gol,'(a,": begin; ",a,"; ",a)') rname, trim(paramkey), trim(mf%filetype); call goPr

    select case ( mf%filetype )
#ifdef with_hdf  
      case ( 'hdf' )
        ! read from hdf file:
        call ReadRecord( mf%mf_hdf, paramkey, t1, t2, nuv, nw, &
                           gridtype, levi, &
                           lli, ll, sp_ll, &
                           status )
        IF_NOTOK_RETURN(status=1)
        ! fill some info values:
        call Init( tmi, paramkey, 'unkown', status )
        call AddHistory( tmi, 'archivekey=='//trim(mf%archivekey), status )
#endif
#ifdef with_grib  
      case ( 'grib' )
        ! read from grib file:
        call ReadRecord( mf%mf_grib, paramkey, t1, t2, nuv, nw, &
                           gridtype, levi, &
                           lli, ll, sp_ll, &
                           ggi, gg, sp_gg, &
                           shi, sh, lnsp_sh, &
                           tmi, status )
        IF_NOTOK_RETURN(status=1)
#endif
#ifdef with_ncep  
      case ( 'ncep-cdc' )
        ! read from ncep file:
        call ReadRecord( mf%mf_ncep_cdc, paramkey, t1, t2, nuv, nw, &
                           gridtype, levi, &
                           lli, ll, sp_ll, &
                           ggi, gg, sp_gg, &
                           shi, sh, lnsp_sh, &
                           tmi, status )
        IF_NOTOK_RETURN(status=1)
#endif
#ifdef with_ncep  
      case ( 'ncep-gfs' )
        ! read from ncep file:
        call ReadRecord( mf%mf_ncep_gfs, paramkey, t1, t2, nuv, nw, &
                           gridtype, levi, &
                           lli, ll, sp_ll, &
                           ggi, gg, sp_gg, &
                           shi, sh, lnsp_sh, &
                           tmi, status )
        IF_NOTOK_RETURN(status=1)
#endif
#ifdef with_prism  
      case ( 'prism' )
        ! receive from oasis coupler:
        !call ReadRecord( mf%mf_prism, paramkey, t1, t2, nuv, gridtype, levi, &
        !                 lli, ll, sp_ll, ggi, gg, sp_gg, shi, sh, lnsp_sh, &
        !                 status )
        call ReadRecord( mf%mf_prism, paramkey, t1, t2, nuv, nw, &
                           gridtype, levi, &
                           lli, ll, sp_ll, &
                           ggi, gg, sp_gg, &
                           shi, sh, lnsp_sh, &
                           tmi, status )
        IF_NOTOK_RETURN(status=1)
        ! fill some info values:
        call Init( tmi, paramkey, 'unkown', status )
        call AddHistory( tmi, 'model==oasis_coupler', status )
#endif
#ifdef with_msc
      case ( 'msc-data' )
        ! read from grib file:
        tmid = t1 + (t2-t1)/2
        call ReadRecord( mf%mf_msc, paramkey, tmid, tmid, nuv, nw, &
                           gridtype, levi, &
                           lli, ll, sp_ll, &
                           ggi, gg, sp_gg, &
                           shi, sh, lnsp_sh, &
                           tmi, status )
        IF_NOTOK_RETURN(status=1)
#endif
      case default
        write (gol,'("unsupported filetype `",a,"`")') trim(mf%filetype); call goErr
        write (gol,'("in ",a)') rname; call goErr; status=1; return
    end select
    
    ! ok
    status = 0

    !write (*,'(a,": end")') name

  end subroutine mf_ReadRecord


  ! ***


!  subroutine mf_ReadEqvLatStuff( mf, t1, t2, eqvlatb, eqvinds, status )
!  
!    use GO, only : TDate
!
!#ifdef with_hdf  
!    use tmm_mf_hdf, only : ReadEqvLatStuff
!#endif
!
!    ! --- in/out -------------------------------
!    
!    type(TMeteoFile), intent(inout)      ::  mf
!    type(TDate), intent(in)              ::  t1, t2
!    real, intent(out)                    ::  eqvlatb(:,:)
!    integer, intent(out)                 ::  eqvinds(:,:,:)
!    integer, intent(out)                 ::  status
!    
!    ! --- const --------------------------------------
!    
!    character(len=*), parameter  ::  rname = mname//'/mf_ReadEqvLatStuff'
!    
!    ! --- begin ---------------------------------
!
!    select case ( mf%filetype )
!#ifdef with_hdf  
!      case ( 'hdf' )
!        call ReadEqvLatStuff( mf%mf_hdf, t1, t2, eqvlatb, eqvinds, status )
!        IF_NOTOK_RETURN(status=1)
!#endif
!      case default
!        write (gol,'("unsupported filetype `",a,"`")') trim(mf%filetype); call goErr
!        write (gol,'("in ",a)') rname; call goErr; status=1; return
!    end select
!    
!    ! ok
!    status = 0
!
!  end subroutine mf_ReadEqvLatStuff
  


  ! ===========================================================
  !
  !   open meteo file for output
  !
  ! ===========================================================
  

  !
  ! Open the meteo file that should contain the field specified by
  ! archivekey, parameter key, time,
  ! or do nothing if the requested file has been opened already.
  !
  ! <archivekey> = <archivetype>:<archivename>
  !
  !    tmpp:od-fc-ml60-glb3x2
  !
  
  subroutine mf_SetupOutput( mf, archivekey, paramkey, tday, t1, t2, &
                                 rcfilename, dir, status )

    use GO, only : goSplitLine
    use GO, only : TrcFile, Init, Done, ReadRc
    use GO, only : TDate

#ifdef with_hdf  
    use tmm_mf_hdf  , only : Init, Get
#endif
    
    ! --- in/out -------------------------------------

    type(TMeteoFile), intent(inout)        ::  mf
    character(len=*), intent(in)           ::  archivekey
    character(len=*), intent(in)           ::  paramkey
    type(TDate), intent(in)                ::  tday, t1, t2
    character(len=*), intent(in)           ::  rcfilename
    character(len=*), intent(in)           ::  dir
    integer, intent(inout)                 ::  status
    
    ! --- const -------------------------------------
    
    character(len=*), parameter ::  rname = mname//'/mf_SetupOutput'
    
    ! --- local -------------------------------------
    
    character(len=10)        ::  archivetype
    character(len=256)       ::  archivename
    
!    character(len=256)       ::  command
!    integer                  ::  year1, month1, day1, hour1
!    integer                  ::  year2, month2, day2, hour2
!    integer                  ::  dth
    type(TrcFile)            ::  infofile

    character(len=256)       ::  archivename2
    character(len=10)        ::  mclass
    character(len=10)        ::  mtype
    character(len=10)        ::  mlevs
    character(len=10)        ::  mgrid
    character(len=10)        ::  filekey
    character(len=16)        ::  treskey
    logical                  ::  with_spm

    ! --- begin -------------------------------------
    
    ! store archive key:
    mf%archivekey = trim(archivekey)
    
    ! split archive key in type and name:
    call goSplitLine( archivekey, archivetype, ':', archivename, status )
    IF_NOTOK_RETURN(status=1)
    
    ! deceide on archive type:
    select case ( archivetype )
    
#ifdef with_hdf
      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      ! daily hdf day files
      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

      case ( 'tm5' )

        ! always hdf files:
        mf%filetype = 'hdf'

        ! setup file:
        call Init( mf%mf_hdf, 'o', dir, archivename//';form=tm5', paramkey, &
                               tday, t1, t2, status )
        IF_NOTOK_RETURN(status=1)
        
        ! store filename:
        mf%filename = mf%mf_hdf%fname
        
        ! extract time range:
        call Get( mf%mf_hdf, status, trange1=mf%t1, trange2=mf%t2 )
        IF_NOTOK_RETURN(status=1)
        
        ! extract paramkeys for fields in file:
        call Get( mf%mf_hdf, status, paramkeys=mf%paramkeys )
        IF_NOTOK_RETURN(status=1)

#endif

      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      ! error ...
      ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      case default

        write (gol,'("unsupported archivetype `",a,"`")') trim(archivetype); call goErr
        write (gol,'("in ",a)') rname; call goErr; status=1; return

    end select

    ! file is opened (or, at least file name is known)
    mf%opened = .true.
    
    ! ok
    status = 0

  end subroutine mf_SetupOutput
  

  ! ***
    

  subroutine mf_WriteRecord_2d( mf, tmi, paramkey, unit, tday, t1, t2, &
                                lli, nuv, ll, status )

    use GO        , only : TDate
    use Grid      , only : TllGridInfo
    use tmm_info  , only : TMeteoInfo
#ifdef with_hdf  
    use tmm_mf_hdf, only : WriteRecord
#endif

    ! --- in/out -------------------------------
    
    type(TMeteoFile), intent(inout)      ::  mf
    type(TMeteoInfo), intent(in)         ::  tmi
    character(len=*), intent(in)         ::  paramkey, unit
    type(TDate), intent(in)              ::  tday, t1, t2
    type(TllGridInfo), intent(in)        ::  lli
    character(len=1), intent(in)         ::  nuv
    real, intent(in)                     ::  ll(:,:)
    integer, intent(out)                 ::  status
    
    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/mf_WriteRecord_2d'
    
    ! --- begin ---------------------------------
    
    select case ( mf%filetype )
#ifdef with_hdf  
      case ( 'hdf' )
        call WriteRecord( mf%mf_hdf, tmi, paramkey, unit, tday, t1, t2, &
                            lli, nuv, ll, status )
        IF_NOTOK_RETURN(status=1)
#endif
      case default
        write (gol,'("unsupported filetype `",a,"`")') trim(mf%filetype); call goErr
        write (gol,'("in ",a)') rname; call goErr; status=1; return
    end select
    
    ! ok
    status = 0

  end subroutine mf_WriteRecord_2d


  ! ***
    

  subroutine mf_WriteRecord_3d( mf, tmi, spname, paramkey, unit, tday, t1, t2, &
                                lli, nuv, levi, nw, ps, ll, status )

    use GO        , only : TDate
    use Grid      , only : TllGridInfo, TLevelInfo
    use tmm_info  , only : TMeteoInfo
#ifdef with_hdf  
    use tmm_mf_hdf, only : WriteRecord
#endif

    ! --- in/out -------------------------------
    
    type(TMeteoFile), intent(inout)      ::  mf
    type(TMeteoInfo), intent(in)         ::  tmi
    character(len=*), intent(in)         ::  spname, paramkey, unit
    type(TDate), intent(in)              ::  tday, t1, t2
    type(TllGridInfo), intent(in)        ::  lli
    character(len=1), intent(in)         ::  nuv
    type(TLevelInfo), intent(in)         ::  levi
    character(len=1), intent(in)         ::  nw
    real, intent(in)                     ::  ps(:,:)
    real, intent(in)                     ::  ll(:,:,:)
    integer, intent(out)                 ::  status
    
    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/mf_WriteRecord_3d'
    
    ! --- begin ---------------------------------
    
    select case ( mf%filetype )
#ifdef with_hdf  
      case ( 'hdf' )
        call WriteRecord( mf%mf_hdf, tmi, spname, paramkey, unit, tday, t1, t2, &
                            lli, nuv, levi, nw, ps, ll, status )
        IF_NOTOK_RETURN(status=1)
#endif
      case default
        write (gol,'("unsupported filetype `",a,"`")') trim(mf%filetype); call goErr
        write (gol,'("in ",a)') rname; call goErr; status=1; return
    end select
    
    ! ok
    status = 0

  end subroutine mf_WriteRecord_3d


end module tmm_mf
